import os.path

import numpy as np
import os
import pickle
from PIL import Image
from torchvision.datasets.vision import VisionDataset
from typing import Any, Callable, Optional, Tuple
import torch


class CIFAR10Embedding(VisionDataset):
    """
    Args:
        root (string): Root directory of dataset where directory
            ``cifar-10-batches-py`` exists or will be saved to if download is set to True.
        train (bool, optional): If True, creates dataset from training set, otherwise
            creates from test set.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
    """
    url = "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
    filename = "cifar-10-python.tar.gz"
    tgz_md5 = 'c58f30108f718f92721af3b95e74349a'
    train_list = [
        ['data_batch_1', 'c99cafc152244af753f735de768cd75f'],
        ['data_batch_2', 'd4bba439e000b95fd0a9bffe97cbabec'],
        ['data_batch_3', '54ebc095f3ab1f0389bbae665268c751'],
        ['data_batch_4', '634d18415352ddfa80567beed471001a'],
        ['data_batch_5', '482c414d41f54cd18b22e5b47cb7c3cb'],
    ]

    test_list = [
        ['test_batch', '40351d587109b95175f43aff81a1287e'],
    ]
    meta = {
        'filename': 'batches.meta',
        'key': 'label_names',
        'md5': '5ff9c542aee3614f3951f8cda6e48888',
    }

    def __init__(
            self,
            root: str,
            embeddings_path: str = None,
            train: bool = True,
            transform: Optional[Callable] = None,
            target_transform: Optional[Callable] = None,
    ) -> None:

        super(CIFAR10Embedding, self).__init__(root, transform=transform,
                                               target_transform=target_transform)

        self.train = train  # training set or test set
        if embeddings_path in ("", '', None, -1):
            embeddings_path = None
        self.embeddings_path = embeddings_path

        if self.train:
            downloaded_list = self.train_list
        else:
            downloaded_list = self.test_list

        self.data: Any = []
        self.targets = []
        self.embeddings = []

        # now load the picked numpy arrays
        for file_name, checksum in downloaded_list:
            file_path = os.path.join(self.root, file_name)
            with open(file_path, 'rb') as f:
                entry = pickle.load(f, encoding='latin1')
                if b'data' in entry:
                    self.data.append(entry[b'data'])
                elif 'data' in entry:
                    self.data.append(entry['data'])
                else:
                    raise Exception("No data entry.")

                if self.embeddings_path is None:
                    if b'embeddings' in entry:
                        self.embeddings.extend(entry[b'embeddings'])
                    elif 'embeddings' in entry:
                        self.embeddings.extend(entry['embeddings'])
                    else:
                        raise Exception("No embeddings")

                if b'labels' in entry:
                    self.targets.extend(entry[b'labels'])
                elif 'labels' in entry:
                    self.targets.extend(entry['labels'])
                elif b'fine_labels' in entry:
                    self.targets.extend(entry[b'fine_labels'])
                elif 'fine_labels' in entry:
                    self.targets.extend(entry['fine_labels'])
                else:
                    raise Exception("No labels in entry.")

        self.data = np.vstack(self.data).reshape(-1, 3, 32, 32)
        self.data = self.data.transpose((0, 2, 3, 1))  # convert to HWC

        self._load_meta()
        if self.embeddings_path is not None:
            self.embeddings = np.load(self.embeddings_path)

    def _load_meta(self) -> None:
        path = os.path.join(self.root, self.meta['filename'])
        with open(path, 'rb') as infile:
            data = pickle.load(infile, encoding='latin1')
            self.classes = data[self.meta['key']]
        self.class_to_idx = {_class: i for i, _class in enumerate(self.classes)}

    def __getitem__(self, index: int) -> Tuple[Any, Any, Any]:
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image, target) where target is index of the target class.
        """
        img, target = self.data[index], self.targets[index]
        embedding = self.embeddings[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        embedding = torch.tensor(embedding)

        return img, target, embedding

    def __len__(self) -> int:
        return len(self.data)

    def extra_repr(self) -> str:
        return "Split: {}".format("Train" if self.train is True else "Test")


class CIFAR10EmbeddingStealing(CIFAR10Embedding):
    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        img, _, embedding = super(CIFAR10EmbeddingStealing, self).__getitem__(
            index)
        return img, embedding


class CIFAR10EmbeddingEval(CIFAR10Embedding):
    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        _, target, embedding = super(CIFAR10EmbeddingEval, self).__getitem__(
            index)
        return embedding, target
